@@ -44,8 +44,6 @@
 #define snprintf _snprintf
 #endif
 
-#define MIN_NONDENORMAL (1.1754944909521339405e-38)
-
 void _orc_profile_init(void);
 
 OrcRandomContext rand_context;
@@ -490,7 +488,7 @@ print_array_val_hex (OrcArray *array, int i, int j)
   }
 }
 
-static orc_uint64
+static void
 print_array_val_float (OrcArray *array, int i, int j)
 {
   void *ptr = ORC_PTR_OFFSET (array->data,
@@ -500,18 +498,15 @@ print_array_val_float (OrcArray *array, int i, int j)
     case 4:
       if (isnan(*(float *)ptr)) {
         printf(" nan %08x", *(orc_uint32 *)ptr);
-        /* This is to get around signaling/non-signaling nans in the output */
-        return (*(orc_uint32 *)ptr) & 0xffbfffff;
       } else {
         printf(" %12.5g", *(float *)ptr);
-        return *(orc_int32 *)ptr;
       }
+      break;
     case 8:
       printf(" %12.5g", *(double *)ptr);
-      return *(orc_int64 *)ptr;
+      break;
     default:
       printf(" ERROR");
-      return -1;
   }
 }
 
@@ -527,10 +522,16 @@ float_compare (OrcArray *array1, OrcArray *array2, int i, int j)
     case 4:
       if (isnan(*(float *)ptr1) && isnan(*(float *)ptr2)) return TRUE;
       if (*(float *)ptr1 == *(float *)ptr2) return TRUE;
-      if (fabs(*(float *)ptr1 - *(float *)ptr2) < MIN_NONDENORMAL) return TRUE;
+      if ((*(float *)ptr1 < 0.0) == (*(float *)ptr2 < 0.0) &&
+          abs((orc_int32)(*(orc_uint32 *)ptr1 - *(orc_uint32 *)ptr2) <= 2))
+        return TRUE;
       return FALSE;
     case 8:
-      /* FIXME */
+      if (isnan(*(double *)ptr1) && isnan(*(double *)ptr2)) return TRUE;
+      if (*(double *)ptr1 == *(double *)ptr2) return TRUE;
+      if ((*(double *)ptr1 < 0.0) == (*(double *)ptr2 < 0.0) &&
+          llabs((orc_int64)(*(orc_uint64 *)ptr1 - *(orc_uint64 *)ptr2) <= 2))
+        return TRUE;
       return FALSE;
   }
   return FALSE;
@@ -724,8 +725,8 @@ orc_test_compare_output_full (OrcProgram *program, int flags)
         for(l=ORC_VAR_D1;l<ORC_VAR_D1+4;l++){
           if (program->vars[l].size > 0) {
             if (flags & ORC_TEST_FLAGS_FLOAT) {
-              a = print_array_val_float (dest_emul[l-ORC_VAR_D1], i, j);
-              b = print_array_val_float (dest_exec[l-ORC_VAR_D1], i, j);
+              print_array_val_float (dest_emul[l-ORC_VAR_D1], i, j);
+              print_array_val_float (dest_exec[l-ORC_VAR_D1], i, j);
               if (!float_compare (dest_emul[l-ORC_VAR_D1], dest_exec[l-ORC_VAR_D1], i, j)) {
                 line_bad = TRUE;
               }
